#include <Foundation/NSObject.h>
#include <Foundation/NSString.h>
#include <Foundation/NSTask.h>
#include <Foundation/NSFileHandle.h>
#include <Foundation/NSException.h>
#include <Foundation/NSNotification.h>
#include <Foundation/NSScanner.h>
#include <Foundation/NSAutoreleasePool.h>
#include <Foundation/NSDebug.h>

#include "GDBSession.h"
#include "GDBResult.h"

#include <libutil.h>
#include <unistd.h>
#include <termios.h>

NSString *GDBSessionInconsistencyException = @"GDBSessionInconsistencyException";
NSString *GDBSessionInternalException = @"GDBSessionInternalException";
NSString *GDBSessionInvalidArgumentException = @"GDBSessionInvalidArgumentException";

#define INPUT_BUFSIZE 1024

@implementation GDBSession

- (id) init
{
  inputBuffer = [[NSMutableString alloc] initWithCapacity:INPUT_BUFSIZE];
  return [super init];
}

- (void) dealloc
{
  /*if (inputPipe)
    RELEASE(inputPipe);
  if (outputPipe)
    RELEASE(outputPipe);*/
  if (sessionInput)
    RELEASE(sessionInput);
  if (sessionOutput)
  {
    [[NSNotificationCenter defaultCenter] 
      removeObserver:self
      name:NSFileHandleDataAvailableNotification
      object:sessionOutput];
    RELEASE(sessionOutput);
  }
  if (sessionTask)
    RELEASE(sessionTask);

  RELEASE(inputBuffer);

  RELEASE(target);

  [super dealloc];
}

- (void) setDelegate:(id)del
{
  delegate = del;
}

- (id) delegate
{
  return delegate;
}

- (void) start
{
  NSTask *task;
  NSNotificationCenter *defaultCenter;
  NSPipe *inputPipe, *outputPipe;

  if (sessionTask)
    [NSException raise:GDBSessionInconsistencyException
      format:@"GDB Session is already started."];

  defaultCenter = [NSNotificationCenter defaultCenter];

  /*openpty(&infmaster, &infslave, NULL, NULL, NULL);
  if (infmaster < 0)
    [NSException raise:GDBSessionInternalException
      format:@"Unable to create pseudo-terminal for inferior."];
  infSlaveName = ttyname(infslave);

  if (!infSlaveName)
    [NSException raise:GDBSessionInternalException
      format:@"Unable to get name of inferior pseudo-terminal."];
  NSDebugLLog(@"GDBSession", @"Using %s for inferior pseudo-terminal.", infSlaveName);
  if (unlockpt(infslave))
    NSDebugLLog(@"GDBSession", @"Unable to unlock slave pty device.");
  if (grantpt(infmaster))
    NSDebugLLog(@"GDBSession", @"Unable to change perms slave pty device.");
  
  slavePty = [[NSFileHandle alloc] initWithFileDescriptor:infmaster 
    closeOnDealloc:YES];
  [defaultCenter addObserver:self
    selector:@selector(receivedInferiorOutput:)
      name:NSFileHandleDataAvailableNotification
      object:slavePty];
  //[slavePty waitForDataInBackgroundAndNotify];
  */

  task = [[NSTask alloc] init];
  [task setLaunchPath:@"gdb"];
  /*if (![task usePseudoTerminal])
  {
    RELEASE(task);
    [NSException raise:GDBSessionInternalException
      format:@"Unable to create psuedo-terminal for GDB session."];
  }*/
  inputPipe = [NSPipe pipe];
  outputPipe = [NSPipe pipe];

  [task setStandardInput:inputPipe];
  [task setStandardOutput:outputPipe];
  
  [task setArguments:
    [NSArray arrayWithObjects:
      [NSString stringWithFormat: @"--tty=%s", "/dev/null"],
      @"--interpreter=mi", nil
    ]];
  [task launch];

  /*struct termios tc;
  if (tcgetattr([sessionInput fileDescriptor], &tc)==0)
  {
    NSDebugLLog(@"GDBSession", @"ICANON=%d, ISIG=%d", tc.c_lflag & ICANON ? 1 : 0,
      tc.c_lflag & ISIG ? 1 : 0);
    tc.c_lflag |= ICANON | ISIG;
    tcsetattr([sessionInput fileDescriptor], TCSANOW, &tc);
  }
  else
    NSDebugLLog(@"GDBSession", @"Unable to change terminal modes for task.");
 */
 /* NSLog(@"sessionOutput=%s, sessionInput=%s, slavePty=%s",
    ttyname([sessionOutput fileDescriptor]),
    ttyname([sessionInput fileDescriptor]),
    ttyname([slavePty fileDescriptor]));
  */
  /*self->sessionInput = RETAIN([task standardInput]); 
  self->sessionOutput = RETAIN([task standardOutput]);*/
  self->sessionInput = RETAIN([inputPipe fileHandleForWriting]);
  self->sessionOutput = RETAIN([outputPipe fileHandleForReading]);
  self->sessionTask = task;

  [defaultCenter addObserver:self
    selector:@selector(receivedTaskOutput:)
    name:NSFileHandleDataAvailableNotification
    object:sessionOutput];

  /* GDB sends a number of out-of band records when we start-up */
  [self waitForInputAndNotify];
}

- (void) terminate
{
  if (!self->sessionTask)
    [NSException raise:GDBSessionInconsistencyException
      format:@"Attempted to end non-existant GDB session."];

  [[NSNotificationCenter defaultCenter] 
    removeObserver:self
    name:NSFileHandleDataAvailableNotification
    object:sessionOutput];
  [sessionTask terminate];
  if (sessionTask)
    DESTROY(sessionTask);
  if (sessionInput)
    DESTROY(sessionInput);
  if (sessionOutput)
    DESTROY(sessionOutput);
  if (slavePty)
    DESTROY(slavePty);

  [inputBuffer deleteCharactersInRange:NSMakeRange(0, [inputBuffer length])];
}

- (void) receivedTaskOutput:(NSNotification*)notification
{
  NSData *data;
  NSString *temp;
  GDBResult* res;

  data = [sessionOutput availableData]; 
  temp = [NSString stringWithCString:(const char*)[data bytes]
    length:[data length]];

  [inputBuffer appendString:temp];
  if ([self parseInputBuffer:&res]) 
    [self waitForInputAndNotify];
  if (res)
  {
    if ([res resultClass]==GDBStoppedResult)
    {
      targetStatus = GDBTargetInterrupted;
    }
  }
}

/*- (void) receivedInferiorOutput:(NSNotification*)notification
{
  NSData* data;
  NSString* temp;

  data = [slavePty availableData];
  temp = [NSString stringWithCString:(const char*)[data bytes]
    length:[data length]];

  //NSLog(@"<><>%@", temp);
  //[slavePty waitForDataInBackgroundAndNotify];
}*/

- (NSString*) target
{
  return self->target;
}

- (void) setTarget:(NSString*)tar
{
  if (![self isRunning])
    [NSException raise:GDBSessionInvalidArgumentException
      format:@"Attempted to set target whilst no session is running."];

  ASSIGN(self->target, tar);
}

- (BOOL) isRunning
{
  return sessionTask!=nil ? YES : NO;
}

- (void) interruptTarget
{
  //NSData *output;
  //struct termios tc;

  //output = [NSData dataWithBytes:"\x03" length:1];
  //[slavePty writeData:output];
  //tcgetattr([sessionInput fileDescriptor], &tc);
  //write([sessionInput fileDescriptor], &(tc.c_cc[VINTR]), 1);
  //printf("INTR=%d\n", tc.c_cc[VINTR]);
  write([sessionInput fileDescriptor], "\x03", 1);
  //kill([sessionTask processIdentifier], SIGINT);
}

- (GDBTargetStatus) targetStatus
{
  return self->targetStatus;
}

- (GDBResult*) executeSynchronousCommand:(NSString*)command arguments:(NSString*)arguments
{
  GDBResult *res = nil;
  NSString *stringData;
  NSData *data;
  BOOL shouldContinue;

  stringData = [NSString stringWithFormat:@"-%@ %@\n",
    command, arguments];
  data = [NSData dataWithBytes:[stringData cString] length:[stringData cStringLength]];
  [sessionInput writeData:data];
  shouldContinue = YES;
  NS_DURING
    while (shouldContinue)
    {
      data = [sessionOutput availableData];
      [inputBuffer appendString:[NSString stringWithCString:[data bytes] length:[data length]]];
      NSDebugLLog(@"GDBSession", @"Attempting to parse input buffer |%@|", inputBuffer);
      shouldContinue = [self parseInputBufferUntilResult:(&res)];
    }
  NS_HANDLER
    NSLog(@"%@", localException);
    [inputBuffer deleteCharactersInRange:NSMakeRange(0, [inputBuffer length])];
    return nil;
  NS_ENDHANDLER;

  if ([res resultClass]==GDBRunningResult)
  {
    /* We can expect out-of band records after the previous result so
       we set our selves up to process it now. */
    NSDebugLLog(@"GDBSession", @"running: remaining inputBuffer:(%d)%@", 
      [inputBuffer length], 
      [inputBuffer dataUsingEncoding:NSUnicodeStringEncoding allowLossyConversion:YES]);
    [self waitForInputAndNotify];
    targetStatus = GDBTargetRunning;
  }

  return res;
}

- (BOOL) parseInputBuffer:(GDBResult**)outOfBand
{
  NSScanner *scanner;
  NSString *parsedLine = nil;
  NSCharacterSet *newlineChars;
  SEL processor;
  GDBStreamIdentifier ident = GDBNilStreamIdentifier;
  BOOL shouldContinue = YES;
  id temp;

  *res = nil;

  scanner = [NSScanner scannerWithString:[NSString stringWithString:inputBuffer]];
  newlineChars = [NSCharacterSet characterSetWithCharactersInString:@"\r\n"];
  [scanner setCharactersToBeSkipped:nil];
  while ([scanner scanUpToCharactersFromSet:newlineChars intoString:&parsedLine])
  {
    if ([scanner isAtEnd])
    {
      NSDebugLLog(@"GDBSession", @"Encountered partial output line with input buffer |%@|",
        inputBuffer);
      break; /* Partial output line should be ignored until we get more data */
    }

    if ([parsedLine rangeOfString:@"(gdb)"].length > 0)
    {
      /* Reached end of result records. Clean out this string and return, leaving anything else
      to be processed when more output comes through (not that there should be anything after this
      unless it is asynchronous) */
      shouldContinue = NO;
    }
    else if ([parsedLine characterAtIndex:0]=='-')
    {
      // Terminal echo of a command we sent. This shouldn't occur here.
    }
    else if ([parsedLine length] > 0)
    {
      processor = [self processorForStreamWithToken:[parsedLine characterAtIndex:0]];
      ident = [self streamIdentifierWithToken:[parsedLine characterAtIndex:0]];
      //if (ident!=GDBNilStreamIdentifier)
      
      //  [NSException raise:GDBSessionInternalException
      //    format:@"Invalid stream identifier:%c", [parsedLine characterAtIndex:0]];
      //else if (ident==GDBResultStreamIdentifier)
      //  NSLog(@"Warning: Shouldn't receive result stream in out-of-band processing.");
      
      IMP processFunction = [self methodForSelector:processor];
      temp = (*processFunction)(self, processor, parsedLine);
      if (ident==GDBOutOfBandStreamIdentifier)
        *res = temp;
    }
    else
      [NSException raise:GDBSessionInternalException 
        format:@"Error whilst parsing GDB MI output."];

    /* Delete the parsed input line from the input buffer */
    if (parsedLine)
      [inputBuffer deleteCharactersInRange:NSMakeRange(0, [parsedLine length])];

    /* Scan the newline characters and remove them from the input buffer */
    if (![scanner scanCharactersFromSet:newlineChars intoString:&parsedLine])
      break; /* Can't possibly continue as no newline characters are there */

    [inputBuffer deleteCharactersInRange:NSMakeRange(0, [parsedLine length])];

    parsedLine = nil;
  }

  return shouldContinue;
}

/**
  * Similar to -parseInputBuffer, but executes synchronously until a result
  * stream is returned.
  */
- (BOOL) parseInputBufferUntilResult:(GDBResult**) result
{
  NSScanner *scanner;
  NSString *parsedLine = nil;
  NSCharacterSet *newlineChars;
  SEL processor;
  GDBStreamIdentifier ident = GDBNilStreamIdentifier;
  id temp = nil;
  BOOL shouldContinue = YES;

  scanner = [NSScanner scannerWithString:[NSString stringWithString:inputBuffer]];
  newlineChars = [NSCharacterSet characterSetWithCharactersInString:@"\r\n"];
  [scanner setCharactersToBeSkipped:nil];
  while (shouldContinue && 
    [scanner scanUpToCharactersFromSet:newlineChars intoString:&parsedLine])
  {
    if ([scanner isAtEnd])
    {
      NSDebugLLog(@"GDBSession", @"Encountered partial output line with input buffer |%@|",
        inputBuffer);
      break; /* Partial output line should be ignored until we get more data */
    }

    if ([parsedLine rangeOfString:@"(gdb)"].length > 0)
    {
      /* Reached end of result records. Clean out this string and return, leaving anything else
      to be processed when more output comes through (not that there should be anything after this
      unless it is asynchronous) */
      NSDebugLLog(@"GDBSession", @"Encountered (gdb) sequence");
      shouldContinue = NO;
    }
    else if ([parsedLine characterAtIndex:0]=='-')
    {
      // Terminal echo of the command we sent. Ignore it.
    }
    else if ([parsedLine length] > 0)
    {
      processor = [self processorForStreamWithToken:[parsedLine characterAtIndex:0]];
      ident = [self streamIdentifierWithToken:[parsedLine characterAtIndex:0]];
      if (ident!=GDBNilStreamIdentifier)
      {
      //  [NSException raise:GDBSessionInternalException
      //    format:@"Invalid stream identifier:%c", [parsedLine characterAtIndex:0]];
        temp = (*objc_msg_lookup(self, processor))(self, processor, parsedLine);
        if (ident==GDBResultStreamIdentifier)
          *result = temp;
      }
    }
    else
      [NSException raise:GDBSessionInternalException 
        format:@"Error whilst parsing GDB MI output."];

    /* Delete the parsed input line from the input buffer */
    if (parsedLine!=nil)
      [inputBuffer deleteCharactersInRange:NSMakeRange(0, [parsedLine length])];

    /* Scan the newline characters and remove them from the input buffer */
    if (![scanner scanCharactersFromSet:newlineChars intoString:&parsedLine])
    {
      //NSLog(@"Trouble scanning newline characters in GDBSession.");
      [NSException raise:GDBSessionInternalException
        format:@"Error trying to parse newline characters in GDB MI input."];
      break; /* Can't possibly continue as no newline characters are there */
    }

    [inputBuffer deleteCharactersInRange:NSMakeRange(0, [parsedLine length])];

    parsedLine = nil;
  }

  return shouldContinue;
}

- (GDBStreamIdentifier) streamIdentifierWithToken:(unichar)token
{
  GDBStreamIdentifier i;
  switch (token)
  {
  case L'^':
    i = GDBResultStreamIdentifier;
    break;
  case L'~':
    i = GDBCLIStreamIdentifier;
    break;
  case L'@':
    i = GDBTargetOutputStreamIdentifier;
    break;
  case L'&':
    i = GDBInternalStreamIdentifier;
    break;
  case L'*':
    i = GDBOutOfBandStreamIdentifier;
    break;
  default:
    i = GDBNilStreamIdentifier;
    break;
  }
  return i;
}

- (SEL) processorForStreamWithToken:(unichar)token
{
  switch (token)
  {
  case L'~':
    return @selector(consoleStreamProcessor:);
  case L'@':
    return @selector(targetStreamProcessor:);
  case L'&':
    return @selector(debugStreamProcessor:);
  case L'*':
    return @selector(outOfBandStreamProcessor:);
  case L'^':
    return @selector(resultStreamProcessor:);
  default:
    return @selector(nilStreamProcessor:);
  }
}

- (id) nilStreamProcessor:(NSString*)obj
{
  //NSLog(obj);
  return self;
}

- (id) consoleStreamProcessor:(NSString*)obj
{
  int temp = 0;
  const char* processString;

  processString = [[obj substringFromIndex:1] cString];
  if (processString[0]=='\0')
    return self;

  //if (processString[0]=='\"')
  //  processString++;

  printf("C>>%s", [[self parseResultString:processString withStartPosition:&temp] cString]);
  return self;
}

- (id) outOfBandStreamProcessor:(NSString*)obj
{
  NSString *processString;
  NSString *resultClass;
  NSDictionary *resultRecords;

  resultClass = nil;
  resultRecords = nil;

  processString = [obj substringFromIndex:1];

  [self parseResultRecord:processString
    class:&resultClass
    records:&resultRecords];

  /* process result;*/
  return self;
}

- (id) resultStreamProcessor:(NSString*)record
{
  GDBResult *result;
  NSDictionary *records;
  NSString *recordClass;
  GDBResultClass cl;
 
  if ([self streamIdentifierWithToken:
    [record characterAtIndex:0]]!=GDBResultStreamIdentifier)
    return nil;

  if (![self parseResultRecord:record
    class:&recordClass
    records:&records])
    return nil;

  cl = [self resultClassWithString:recordClass];
  result = [[GDBResult alloc] 
    initWithResultClass:cl records:records];

  return AUTORELEASE(result);
}

- (BOOL) parseResultRecord:(NSString*)rec
  class:(NSString**)recordClass
  records:(NSDictionary**)records
{
  const char *record;
  NSMutableString *temp;
  int pos = 0;  
 
  *recordClass = nil;
  *records = nil;

  record = [rec cString];
  if (record==NULL || strlen(record)<=0)
    return NO;

  record++; // skip stream identifier

  temp = [NSMutableString stringWithCapacity:10];
  while (*record!='\0' && *record!=',')
  {
    [temp appendFormat:@"%c", *record];
    record++;
  }
  
  /* If we have reached the end of the line and scanned in *something*
     then we know that we have received a result record without a
     dictionary, e.g. "^done\n", which is valid. */

  if ([temp length]<= 0)
    return NO;

  if (*record=='\0')
  {
    *recordClass = temp;
    return YES;
  }
  else if (*record!=',')
    return NO;
  
  // We know there is a comma so there should be more to come.
  record++;

  *records = [self parseResultDictionary:record withStartPosition:&pos];
  if (*records==nil)
    return NO;

  *recordClass = temp;

  return (*recordClass) ? YES : NO;
}

- (NSDictionary*) parseResultDictionary:(const char*)record withStartPosition:(int*)pos
{
  int length, temp;
  NSMutableString *name;
  id value;
  NSMutableDictionary* dict;

  dict = [NSMutableDictionary dictionaryWithCapacity:10];

  length = strlen(record);

  while (*pos < length)
  {
    temp = *pos;
    while (*pos < length && record[*pos]!='=' && record[*pos]!=',')
    {
      (*pos)++;
    }
    name = [NSString stringWithCString:(record + temp) length:(*pos - temp)];
    
    if (record[*pos]=='=')
    {
      (*pos)++;
      switch (record[*pos])
      {
      case '{':
        (*pos)++;
        value = [self parseResultDictionary:record withStartPosition:pos];
        break;
      case '[':
        (*pos)++;
        value = [self parseResultArray:record withStartPosition:pos];
        break;
      case '\"':
        //(*pos)++;
        value = [self parseResultString:record withStartPosition:pos];
        break;
      default:
        return nil;
      }
    }
    if (value==nil)
      return nil;
    
    [dict setObject:value forKey:name];

    if (record[*pos]==',')
      (*pos)++;
    else if (record[*pos]=='}')
    {
      (*pos)++;
      break;
    }
    else if (record[*pos]!='\0')
      return nil;
  }
  return dict;
}

- (NSArray*) parseResultArray:(const char*)record withStartPosition:(int*)pos
{
  int length;
  id value;
  NSMutableArray *ar;

  length = strlen(record);
  ar = [NSMutableArray arrayWithCapacity:10];

  while (*pos < length)
  {
    switch (record[*pos])
    {
    case '{':
      (*pos)++;
      value = [self parseResultDictionary:record withStartPosition:pos];
      break;
    case '[':
      (*pos)++;
      value = [self parseResultArray:record withStartPosition:pos];
      break;
    case '\"':
      (*pos)++;
      value = [self parseResultArray:record withStartPosition:pos];
      break;
    default:
      value = nil;
      break;
    }
  
    if (value==nil)
      return nil;
    if (record[*pos]==']')
    {
      (*pos)++;
      break;
    }
    else if (record[*pos]!=',')
      return nil;

    [ar addObject:value];
    (*pos)++;
  }
  return ar; 
}

- (NSString*) parseResultString:(const char*)record withStartPosition:(int*)pos
{
  NSMutableString *string;
  int temp;
  BOOL inQuote;

  inQuote = record[*pos]=='\"' ? YES : NO;
  if (inQuote)
    *pos += 1;

  string = [NSMutableString stringWithCapacity:10];
  while (record[*pos]!='\0')
  {
    switch (record[*pos])
    {
    case '\\':
      if (!inQuote)
      {
        temp = '\\';
        break;
      }
      if (record[++(*(pos))]=='\0')
        return nil;
      switch (record[*pos])
      {
      case 'n':
        temp = '\n';
        break;
      case '\\':
        temp = '\\';
        break;
      case 't':
        temp = '\t';
        break;
      case '\"':
        temp = '\"';
        break;
      case '\'':
        temp = '\'';
        break;
      case 'r':
        temp = '\r';
      default:
        temp = ' ';
        break;
      }
      break;
    case '"':
      if (inQuote)
      {
        (*pos)++;
        return string;
      }
      break;
    default:
      temp = record[*pos];
    }
    if (temp < 0)
      continue;
    [string appendFormat:@"%c", temp];
    (*pos)++;
  }
  return string;
}

- (GDBResultClass) resultClassWithString:(NSString*)aString
{
  if ([aString isEqualToString:@"done"])
    return GDBDoneResult;
  else if ([aString isEqualToString:@"running"])
    return GDBRunningResult;
  else if ([aString isEqualToString:@"connected"])
    return GDBConnectedResult;
  else if ([aString isEqualToString:@"error"])
    return GDBErrorResult;
  else if ([aString isEqualToString:@"exit"])
    return GDBExitResult;
  else if ([aString isEqualToString:@"stopped"])
    return GDBStoppedResult;
  else
    return GDBUnknownResult;
}

- (void) waitForInputAndNotify
{

  [sessionOutput waitForDataInBackgroundAndNotify];
}

@end
